# -*- coding: utf-8 -*-
"""
    pip_services3_container.config.ContainerConfigReader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Container configuration reader implementation.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.errors import ConfigException
from pip_services3_components.config import JsonConfigReader
from pip_services3_components.config import YamlConfigReader
from .ContainerConfig import ContainerConfig


class ContainerConfigReader():
    """
    Helper class that reads container configuration from JSON or YAML file.
    """

    @staticmethod
    def read_from_file(correlation_id, path, parameters):
        """
        Reads container configuration from JSON or YAML file.
        The type of the file is determined by file extension.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param path: a path to component configuration file.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: the read container configuration
        """
        if path == None:
            raise ConfigException(correlation_id, "NO_PATH", "Missing config file path")

        index = path.rfind('.')
        ext = path[index + 1:].lower() if index > 0 else ''

        if ext == "json":
            return ContainerConfigReader.read_from_json_file(correlation_id, path, parameters)
        elif ext == "yaml" or ext == "yml":
            return ContainerConfigReader.read_from_yaml_file(correlation_id, path, parameters)

        # By default read as JSON
        return ContainerConfigReader.read_from_json_file(correlation_id, path, parameters)

    @staticmethod
    def read_from_json_file(correlation_id, path, parameters):
        """
        Reads container configuration from JSON file.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param path: a path to component configuration file.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: the read container configuration
        """
        config = JsonConfigReader.read_config(correlation_id, path, parameters)
        return ContainerConfig.from_config(config)

    @staticmethod
    def read_from_yaml_file(correlation_id, path, parameters):
        """
        Reads container configuration from YAML file.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param path: a path to component configuration file.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: the read container configuration
        """
        config = YamlConfigReader.read_config(correlation_id, path, parameters)
        return ContainerConfig.from_config(config)




# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.LinkReferencesDecorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Link references decorator implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.refer import IReferences
from pip_services3_commons.refer import Referencer
from pip_services3_commons.run import IOpenable

from .ReferencesDecorator import ReferencesDecorator

class LinkReferencesDecorator(ReferencesDecorator, IOpenable):
    """
    References decorator that automatically sets references to newly added components
    that implement IReferenceable interface and unsets references
    from removed components that implement IUnreferenceable interface.
    """
    _opened = False

    def __init__(self, base_references, parent_references):
        """
        Creates a new instance of the decorator.

        :param base_references: the next references or decorator in the chain.

        :param parent_references: the decorator at the top of the chain.
        """
        super(LinkReferencesDecorator, self).__init__(base_references, parent_references)


    def is_opened(self):
        """
        Checks if the component is opened.

        :return: true if the component has been opened and false otherwise.
        """
        return self._opened

    def open(self, correlation_id):
        """
        Opens the component.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        if not self._opened:
            components = self.get_all()
            Referencer.set_references(self.parent_references, components)
            self._opened = True

    def close(self, correlation_id):
        """
        Closes component and frees used resources.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        if self._opened:
            components = self.get_all()
            Referencer.unset_references(components)
            self._opened = False


    def put(self, locator = None, component = None):
        """
        Puts a new reference into this reference map.

        :param locator: a locator to find the reference by.

        :param component: a component reference to be added.
        """
        super(LinkReferencesDecorator, self).put(locator, component)

        if self._opened:
            Referencer.set_references_for_one(self.parent_references, component)


    def remove(self, locator):
        """
        Removes a previously added reference that matches specified locator.
        If many references match the locator, it removes only the first one.
        When all references shall be removed, use [[removeAll]] method instead.

        :param locator: a locator to remove reference

        :return: the removed component reference.
        """
        component = super(LinkReferencesDecorator, self).remove(locator)

        if self._opened:
            Referencer.unset_references_for_one(component)

        return component


    def remove_all(self, locator):
        """
        Removes all component references that match the specified locator.

        :param locator: the locator to remove references by.

        :return: a list, containing all removed references.
        """
        components = super(LinkReferencesDecorator, self).remove_all(locator)

        if self._opened:
            Referencer.unset_references(components)

        return components



# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.ContainerReferences
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Container references implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config import IConfigurable
from pip_services3_components.build import CreateException
from pip_services3_commons.refer import References
from pip_services3_commons.refer import ReferenceException
from pip_services3_commons.refer import IReferenceable
from pip_services3_commons.reflect import TypeReflector

from ..config.ComponentConfig import ComponentConfig
from ..config.ContainerConfig import ContainerConfig
from .ManagedReferences import ManagedReferences

class ContainerReferences(ManagedReferences):
    """
    Container managed references that can be created from container configuration.
    """
    def __init__(self):
        super(ContainerReferences, self).__init__()

    def put_from_config(self, config):
        """
        Puts components into the references from container configuration.

        :param config: a container configuration with information of components to be added.
        """
        for component_config in config:
            component = None
            locator = None

            try:
                # Create component dynamically
                if component_config.type != None:
                    locator = component_config.type
                    component = TypeReflector.create_instance_by_descriptor(component_config.type)
                # Or create component statically
                elif component_config.descriptor != None:
                    locator = component_config.descriptor
                    factory = self._builder.find_factory(locator)
                    component = self._builder.create(locator, factory)
                    if component == None:
                        raise ReferenceException(None, locator)
                    locator = self._builder.clarify_locator(locator, factory)

                # Check that component was created
                if component == None:
                    raise CreateException("CANNOT_CREATE_COMPONENT", "Cannot create component") \
                        .with_details("config", config)

                # Add component to the list
                self._references.put(locator, component)

                # Configure component
                if isinstance(component, IConfigurable):
                    component.configure(component_config.config)

                # Set references to factories
                if isinstance(component, IReferenceable):
                    component.set_references(self)

            except Exception as ex:
                raise ReferenceException(None, locator).with_cause(ex)



# -*- coding: utf-8 -*-
"""
    pip_services3_container.Container
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Container implementation.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import traceback

from pip_services3_commons.config import IConfigurable
from pip_services3_components.info import _DefaultInfoFactory
from pip_services3_components.log import NullLogger
from pip_services3_components.log import CompositeLogger
from pip_services3_commons.errors import InvalidStateException
from pip_services3_commons.refer import Descriptor
from pip_services3_commons.refer import IReferenceable
from pip_services3_commons.refer import IUnreferenceable
from pip_services3_commons.refer import Referencer
from pip_services3_commons.run import IOpenable
from pip_services3_commons.run import Opener
from pip_services3_commons.run import Closer

from .build.DefaultContainerFactory import DefaultContainerFactory
from pip_services3_components.info.ContextInfo import ContextInfo
from pip_services3_components.info._DefaultInfoFactory import DefaultInfoFactory
from .config.ContainerConfigReader import ContainerConfigReader
from .refer.ContainerReferences import ContainerReferences
from .config.ContainerConfig import ContainerConfig


class Container(IConfigurable, IReferenceable, IUnreferenceable, IOpenable):
    """
    Inversion of control (IoC) container that creates components and manages their lifecycle.

    The container is driven by configuration, that usually stored in JSON or YAML file.
    The configuration contains a list of components identified by type or locator, followed by component configuration.

    On container start it performs the following actions:
        - Creates components using their types or calls registered factories to create components using their locators
        - Configures components that implement IConfigurable interface and passes them their configuration parameters
        - Sets references to components that implement IReferenceable interface and passes them references of all components in the container
        - Opens components that implement IOpenable interface

    On container stop actions are performed in reversed order:
        - Closes components that implement ICloseable interface
        - Unsets references in components that implement IUnreferenceable interface
        - Destroys components in the container.

    The component configuration can be parameterized by dynamic values. That allows specialized containers
    to inject parameters from command line or from environment variables.

    The container automatically creates a ContextInfo component that carries detail information
    about the container and makes it available for other components.

    ### Configuration parameters ###

        - name: 					the context (container or process) name
        - description: 		   	human-readable description of the context
        - properties: 			    entire section of additional descriptive properties
        - ...

        Example:
            ======= config.yaml ========
            - descriptor: mygroup:mycomponent1:default:default:1.0
            param1: 123
            param2: ABC

            - type: mycomponent2,mypackage
            param1: 321
            param2: XYZ
            ============================

            container = Container()
            container.add_factory(MyComponentFactory())

            parameters = ConfigParams.from_value(os.env)
            container.read_config_from_file("123", "./config/config.yml", parameters)

            container.open("123")
            print "Container is opened"
            ...
            container.close("123")
            print "Container is closed"
    """
    _logger = None
    _info = None
    _config = None
    _factories = None
    _references = None

    def __init__(self, name=None, description=None):
        """
        Creates a new instance of the container.

        :param name: (optional) a container name (accessible via ContextInfo)

        :param description: (optional) a container description (accessible via ContextInfo)
        """
        self._logger = NullLogger()
        self._info = ContextInfo(name, description)
        self._factories = DefaultContainerFactory()

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._config = ContainerConfig.from_config(config)

    def read_config_from_file(self, correlation_id, path, parameters):
        """
        Reads container configuration from JSON or YAML file and parameterizes it with given values.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param path: a path to configuration file

        :param parameters: values to parameters the configuration or null to skip parameterization.
        """
        self._config = ContainerConfigReader.read_from_file(correlation_id, path, parameters)
        self._logger.trace(correlation_id, self._config.__str__())

    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        pass

    def unset_references(self):
        """
        Unsets (clears) previously set references to dependent components.
        """
        pass

    def _init_references(self, references):
        # Override in base classes
        existingInfo = references.get_one_optional(_DefaultInfoFactory.ContextInfoDescriptor)
        if existingInfo is None:
            references.put(_DefaultInfoFactory.ContextInfoDescriptor, self._info)
        else:
            self._info = existingInfo
        references.put(DefaultContainerFactory.DefaultContainerFactoryDescriptor, self._factories)

    def add_factories(self, factory):
        """
        Adds a factory to the container. The factory is used to create components
        added to the container by their locators (descriptors).

        :param factory: a component factory to be added.
        """
        self._factories.add(factory)

    def is_opened(self):
        """
        Checks if the component is opened.

        :return: true if the component has been opened and false otherwise.
        """
        return self._references is not None

    def open(self, correlation_id):
        """
        Opens the component.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        if self._references is not None:
            raise InvalidStateException(correlation_id, "ALREADY_OPENED", "Container was already opened")

        try:
            self._logger.trace(correlation_id, "Starting container.")

            # Create references with configured components
            self._references = ContainerReferences()
            self._init_references(self._references)
            self._references.put_from_config(self._config)
            self.set_references(self._references)

            # Get custom description if available
            info_descriptor = Descriptor("*", "context-info", "*", "*", "*")
            self._info = self._references.get_one_optional(info_descriptor)

            # Reference and open components
            self._references.open(correlation_id)

            # Get reference to logger
            self._logger = CompositeLogger(self._references)
            self._logger.info(correlation_id, "Container " + self._info.name + " started.")
        except Exception as ex:
            self._logger.error(correlation_id, ex, "Failed to start container")
            traceback.print_exc()
            raise ex

    def close(self, correlation_id):
        """
        Closes component and frees used resources.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        if self._references is None:
            return

        try:
            self._logger.trace(correlation_id, "Stopping " + self._info.name + " container")

            # Unset references for child container
            self.unset_references()

            # Close and deference components
            self._references.close(correlation_id)
            self._references = None

            self._logger.info(correlation_id, "Container " + self._info.name + " stopped")
        except Exception as ex:
            self._logger.error(correlation_id, ex, "Failed to stop container")
            raise ex



# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.ReferencesDecorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    References decorator implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.refer import IReferences

class ReferencesDecorator(IReferences):
    """
    Chainable decorator for IReferences that allows to inject additional capabilities
    such as automatic component creation, automatic registration and opening.
    """
    base_references = None
    parent_references = None

    def __init__(self, base_references, parent_references):
        """
        Creates a new instance of the decorator.

        :param base_references: the next references or decorator in the chain.

        :param parent_references: the decorator at the top of the chain.
        """
        self.base_references = base_references if base_references != None else parent_references
        self.parent_references = parent_references if parent_references != None else base_references


    def put(self, locator = None, reference = None):
        """
        Puts a new reference into this reference map.

        :param locator: a locator to find the reference by.

        :param component: a component reference to be added.
        """
        self.base_references.put(locator, reference)

    def remove(self, locator):
        """
        Removes a previously added reference that matches specified locator.
        If many references match the locator, it removes only the first one.
        When all references shall be removed, use [[removeAll]] method instead.

        :param locator: a locator to remove reference

        :return: the removed component reference.
        """
        return self.base_references.remove(locator)

    def remove_all(self, locator):
        """
        Removes all component references that match the specified locator.

        :param locator: the locator to remove references by.

        :return: a list, containing all removed references.
        """
        return self.base_references.remove_all(locator)

    def get_all_locators(self):
        """
        Gets locators for all registered component references in this reference map.

        :return: a list with component locators.
        """
        return self.base_references.get_all_locators()

    def get_all(self):
        """
        Gets all component references registered in this reference map.

        :return: a list with component references.
        """
        return self.base_references.get_all()
        
    def get_one_optional(self, locator):
        """
        Gets an optional component reference that matches specified locator.

        :param locator: the locator to find references by.

        :return: a matching component reference or null if nothing was found.
        """
        try:
            components = self.find(locator, False)
            return components[0] if len(components) > 0 else None
        except Exception as ex:
            return None

    def get_one_required(self, locator):
        """
        Gets a required component reference that matches specified locator.

        :param locator: the locator to find a reference by.

        :return: a matching component reference.

        :raises: a [[ReferenceException]] when no references found.
        """
        components = self.find(locator, True)
        return components[0] if len(components) > 0 else None

    def get_optional(self, locator):
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find references by.

        :return: a list with matching component references or empty list if nothing was found.
        """
        try:
            return self.find(locator, False)
        except Exception as ex:
            return []

    def get_required(self, locator):
        """
        Gets all component references that match specified locator.
        At least one component reference must be present. If it doesn't the method throws an error.

        :param locator: the locator to find references by.

        :return: a list with matching component references.

        :raises: a [[ReferenceException]] when no references found.
        """
        return self.find(locator, True)

    def find(self, locator, required):
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find a reference by.

        :param required: forces to raise an exception if no reference is found.

        :return: a list with matching component references.

        :raises: a [[ReferenceException]] when required is set to true but no references found.
        """
        return self.base_references.find(locator, required)



# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.BuildReferencesDecorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Build references decorator implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.refer import IReferences
from pip_services3_commons.refer import Descriptor
from pip_services3_commons.refer import ReferenceException
from pip_services3_components.build import IFactory

from .ReferencesDecorator import ReferencesDecorator

class BuildReferencesDecorator(ReferencesDecorator):
    """
    References decorator that automatically creates missing components using
    available component factories upon component retrival.
    """
    def __init__(self, base_references, parent_references):
        """
        Creates a new instance of the decorator.

        :param base_references: the next references or decorator in the chain.

        :param parent_references: the decorator at the top of the chain.
        """
        super(BuildReferencesDecorator, self).__init__(base_references, parent_references)


    def find_factory(self, locator):
        """
        Finds a factory capable creating component by given descriptor
        from the components registered in the references.

        :param locator: a locator of component to be created.

        :return: found factory or null if factory was not found.
        """
        components = self.get_all()
        for component in components:
            if isinstance(component, IFactory):
                if component.can_create(locator) != None:
                    return component
        return None


    def create(self, locator, factory):
        """
        Creates a component identified by given locator.

        :param locator: a locator to identify component to be created.

        :param factory: a factory that shall create the component.

        :return: the created component.
        """
        if factory == None:
            return None

        try:
            # Create component
            return factory.create(locator)
        except Exception as ex:
            return None


    def clarify_locator(self, locator, factory):
        """
        Clarifies a component locator by merging two descriptors into one to replace missing fields.
        That allows to get a more complete descriptor that includes all possible fields.

        :param locator: a component locator to clarify.

        :param factory: a factory that shall create the component.

        :return: clarified component descriptor (locator)
        """
        if factory == None:
            return locator
        if not isinstance(locator, Descriptor):
            return locator
        
        another_locator = factory.can_create(locator)
        if another_locator == None:
            return locator
        if not isinstance(another_locator, Descriptor):
            return locator
        
        descriptor = locator
        another_descriptor = another_locator
        
        return Descriptor(
            descriptor.get_group() if descriptor.get_group() != None else another_descriptor.get_group(),
            descriptor.get_type() if descriptor.get_type() != None else another_descriptor.get_type(),
            descriptor.get_kind() if descriptor.get_kind() != None else another_descriptor.get_kind(),
            descriptor.get_name() if descriptor.get_name() != None else another_descriptor.get_name(),
            descriptor.get_version() if descriptor.get_version() != None else another_descriptor.get_version()
        )


    def find(self, locator, required):
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find a reference by.

        :param required: forces to raise an exception if no reference is found.

        :return: a list with matching component references.
        """
        components = super(BuildReferencesDecorator, self).find(locator, False)

        # Try to create component
        if len(components) == 0:
            factory = self.find_factory(locator)
            component = self.create(locator, factory)
            if component != None:
                try:
                    locator = self.clarify_locator(locator, factory)
                    self.parent_references.put(locator, component)
                    components.append(component)
                except Exception as ex:
                    # Ignore exception
                    pass

        # Throw exception is no required components found
        if required and len(components) == 0:
            raise ReferenceException(None, locator)

        return components



# -*- coding: utf-8 -*-
"""
    pip_services3_container.build.DefaultContainerFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Default container factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.refer import Descriptor
from pip_services3_components.build import CompositeFactory
from pip_services3_components.log import DefaultLoggerFactory
from pip_services3_components.count import DefaultCountersFactory
from pip_services3_components.config import DefaultConfigReaderFactory
from pip_services3_components.cache import DefaultCacheFactory
from pip_services3_components.auth import DefaultCredentialStoreFactory
from pip_services3_components.connect import DefaultDiscoveryFactory
from pip_services3_components.info._DefaultInfoFactory import DefaultInfoFactory

class DefaultContainerFactory(CompositeFactory):
    """
    Creates default container components (loggers, counters, caches, locks, etc.) by their descriptors.
    """
    DefaultContainerFactoryDescriptor = Descriptor(
        "pip-services", "factory", "container", "default", "1.0"
    )

    def __init__(self, *factories):
        """
        Create a new instance of the factory and sets nested factories.

        :param factories: a list of nested factories
        """
        super(DefaultContainerFactory, self).__init__(factories)
        self.add(DefaultInfoFactory())
        self.add(DefaultLoggerFactory())
        self.add(DefaultCountersFactory())
        self.add(DefaultConfigReaderFactory())
        self.add(DefaultCacheFactory())
        self.add(DefaultCredentialStoreFactory())
        self.add(DefaultDiscoveryFactory())
        # TODO Add DefaultTestFactory




# -*- coding: utf-8 -*-
"""
    pip_services3_container.config.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Container configuration serves as a recipe for instantiating and
    configuring components inside the container.

    External configurations (stored as YAML or JSON) are passed to the container
    and define the structure of objects that need to be recreated in the container.
    Objects can be defined in two ways:
        - using descriptors (using which registered factories can recreate the object)
        - using hard-coded types (objects are recreated directly, based on their type, bypassing factories).

    In addition, various configurations are stored for each object. The container recreates the
    objects and, if they implement the IConfigurable interface, passes them their configurations.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'ComponentConfig', 'ContainerConfig', 'ContainerConfigReader'
]

from .ComponentConfig import ComponentConfig
from .ContainerConfig import ContainerConfig
from .ContainerConfigReader import ContainerConfigReader



# -*- coding: utf-8 -*-
"""
    pip_services3_container.build.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    The default container factory provides generic functionality on-demand, such as logging and performance monitoring.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = ['DefaultContainerFactory']

from .DefaultContainerFactory import DefaultContainerFactory



"""
Pip.Services Container
----------------------

Pip.Services is an open-source library of basic microservices.
pip_services3_container provides IoC container implementation.

Links
`````

* `website <http://github.com/pip-services/pip-services>`_
* `development version <http://github.com/pip-services3-python/pip-services3-container-python>`

"""

from setuptools import setup
from setuptools import find_packages
 
setup(
    name='pip_services3_container',
    version='3.1.1',
    url='http://github.com/pip-services3-python/pip-services3-container-python',
    license='MIT',
    author='Conceptual Vision Consulting LLC',
    author_email='seroukhov@gmail.com',
    description='IoC container for Pip.Services in Python',
    long_description=__doc__,
    packages=find_packages(exclude=['config', 'data', 'examples', 'test']),
    include_package_data=True,
    zip_safe=True,
    platforms='any',
    install_requires=[
        'iso8601', 'PyYAML', 'pytest', 'pytz', 'pystache', 'pybars3', 'numpy', 'pip_services3_commons', 'pip_services3_components'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]    
)